## The GENERATE Tool is created by NA-ACCORD for local address geocoding and 
## submission of area-Level indicators

# Note: 
  # Before running this script, please read the "User Guideline.doc" file
  # To run the function degauss_run() in this script, one should firstly install "Docker Desktop" on a Windows or macOS system, or "Docker Engine" on a Linux system
  # Go to "https://docs.docker.com/desktop/" for "Docker Desktop" or "Docker Engine" installation

# Set the Working Directory
# setwd("YOUR/PATH/HERE")

library(dplyr, quietly = TRUE)
library(dht, quietly = TRUE)
library(lubridate)
library(tidyr)
library(tidyverse)

# Requirements of the input address csv file:
  # Should be named as "Participant_Addresses"
  # Should have separate columns for the followings:
    # 1st column: cohortPatientId
    # 2nd column: Street Number
    # 3rd column: Street Name
    # 4th column: Additional Identifiers, such as unit number / apartment number, can leave blank
    # 5th column: City
    # 6th column: State (in postal abbreviation format, 2 capital letters (e.g., CA for California))
    # 7th column: ZIP
    # 8th column: Month (the month of the date when the address was recorded, should be in numbers (i.e. "7" for July)
    # 9th column: Day (the day of the date when the address was recorded, should be in numbers, (i.e. "24" for 24th))
    # 10th column: Year (the year of the date when the address was recorded, should be in numbers)
# Important: 
  # The csv file should have all these 10 columns
  # The name of each column should be exactly what's described above (i.e., cohortPatientId, Street Number, Street Name, Additional Identifier, etc.)
  # The names of the columns should be on the first row


# Import the input csv file
input_address <- read.csv("Participant_Addresses.csv") 

# Remove all duplicated addresses
input_address <- input_address %>% 
  distinct(cohortPatientId, Street.Number, Street.Name, Additional.Identifiers, 
           City, State, ZIP.Code, Month, Day, Year, .keep_all = TRUE)

# Remove the addresses before year 2012
#input_address <- input_address |> filter(Year>=2012)
# If the date is before 2012, the start date will be recorded as 01/01/2012
# So we no longer remove addresses before year 2012

# Remove the addresses after year 2022
################ Note ##################
# We set it to 2022 because the latest data for now is by year 2022.
# This should be updated accordingly in the future.
input_address <- input_address |> filter(Year<=2022)


# we want to have each year recorded for participants' address
# for participants with multiple address, startdate is the address recorded and end date is the day before next address record date
# for participants with single address recorded or the last address among the multiple address, enddate is the date for the last day of the protocol year
# then we want to separate each year in the duration between startdate and enddate to get census information for each year
input_address$record_date <- as.Date(paste(input_address$Year, input_address$Month, input_address$Day, sep = "-"))

input_address <- input_address %>%
  arrange(cohortPatientId, record_date) %>% 
  group_by(cohortPatientId) %>%
  mutate(
    startdate = if_else(year(record_date) < 2012, as.Date("2012-01-01"), record_date),
    enddate = lead(record_date) - 1
  )  %>%
  ungroup()
# For patients with a single address or the last address among multiple addresses, set enddate to the last day of current protocol year
input_address$enddate[is.na(input_address$enddate)] <- as.Date("2022-12-31")

# write a function to expand rows by year
by_year <- function(start, end) {
  seq.Date(start, end, by = "year")
}

expand_years <- function(df) {
  df %>%
    mutate(year = map2(startdate, enddate, by_year)) %>%
    unnest(cols = c(year)) %>%
    mutate(
      start = if_else(year == startdate, startdate, floor_date(year, "year")),
      end = if_else(year == enddate, enddate, ceiling_date(year, "year") - days(1))
    ) %>%
    group_by(cohortPatientId) %>%
    mutate(lead_start = lead(start)) %>%
    mutate(end = if_else(!is.na(lead_start) & lead_start < end, lead_start - days(1), end)) %>%
    ungroup() %>%
    select(-startdate, -enddate, -year, -lead_start)
}


# expand rows by year
input_address <- expand_years(input_address)

# check for (end!=lead(start)-1) to see if still need to insert rows after expansion
insert_row <- input_address %>%
  group_by(cohortPatientId) %>%
  mutate(
    lead_start = lead(start),
    insert = !is.na(lead_start) & end != lead_start - days(1)
  ) %>%
  ungroup() %>%
  filter(insert) %>%
  mutate(end = lead_start - days(1)) %>%
  mutate(start = floor_date(end, "year")) %>%
  select(-lead_start, -insert)

# Combine the original expanded_df with the inserted rows
input_address <- bind_rows(input_address, insert_row) %>%
  arrange(cohortPatientId, start)

#change Year in input address to match the year of start date;
input_address <-input_address %>% mutate(Year=year(start))

# In case the input ZIP codes do not all have 5 digits, fill those with fewer than 5 digits with initial 0s
filled_zip <- sprintf("%05d", input_address$ZIP.Code)
# Add this filled_zip as a new column to the input_address dataframe
input_address$filled_zip <- filled_zip

# Combine the columns together
input_address$combined <- paste(input_address$Street.Number, 
                                tolower(input_address$Street.Name), 
                                #input_address$Additional.Identifiers, 
                                tolower(input_address$City), 
                                tolower(input_address$State),
                                input_address$filled_zip,
                                sep = " ")


# Clean the input addresses, remove non-alphanumeric characters and excess white space
cleaned_address <- dht::clean_address(input_address$combined)


# Obtain the number of rows of the input csv file, and this is the number of addresses
num_rows <- nrow(input_address) 


# Create a tibble for the addresses, and create unique identifier for each address
addresses <-  tibble::tibble(
  cohortPatientId = input_address$cohortPatientId,
  address = cleaned_address #Make sure every word is separated by only one space
)

write.csv(addresses, file = "adrs_geocoded.csv", row.names = FALSE)

# Geocoding addresses with non-parsed addresses
adrs_geocoded <- 
  degauss_run(addresses, "geocoder", "3.2.1", quiet = FALSE)

# Obtain the proportion of precisely geocoded addresses
precise_proportion <- mean(adrs_geocoded$geocode_result == "geocoded")
# Obtain the proportion of imprecisely geocoded addresses
imprecise_proportion <- (1-precise_proportion)

# Obtain which addresses is imprecisely geocoded (the ones without geocoded latitude or longtitude)
imprecise_geocoded <- which (!(adrs_geocoded$geocode_result == "geocoded"))



###### Check states
# Clean the input state - remove excess whitespace (this is for the checking functions of state later)
cleaned_state <- dht::clean_address(input_address$State)

# Check whether the matched state is the same as the state provided. Return the proportion of correctly matched states
state_check <- function(state, matched_state) {
  check_state <- state ==  matched_state
  proportion_matched <- mean(check_state)
  return(proportion_matched)
}
state_matched_proportion <- state_check(tolower(cleaned_state), tolower(adrs_geocoded$matched_state))

# Obtain TRUE/FALSE indicating whether the geocoded state matched the input state
state_match_status <- function(state, matched_state) {
  check_state <- state ==  matched_state
  return(check_state)
}
state_matching <- state_match_status(tolower(cleaned_state), tolower(adrs_geocoded$matched_state))

# Obtain which addresses have unmatched state
false_state <- which(!state_matching)



###### Check cities
# Clean the input city - remove excess whitespace (this is for the checking functions of city later)
cleaned_city <- dht::clean_address(input_address$City)

# Check whether the matched city is the same as the city provided. Return the proportion of correctly matched city
city_check <- function(city, matched_city) {
  check_city <- city ==  matched_city
  proportion_matched <- mean(check_city)
  return(proportion_matched)
}
city_matched_proportion <- city_check(tolower(cleaned_city), tolower(adrs_geocoded$matched_city))

# Obtain TRUE/FALSE indicating whether the geocoded city matched the input city
city_match_status <- function(city, matched_city) {
  check_city <- city ==  matched_city
  return(check_city)
}
city_matching <- city_match_status(tolower(cleaned_city), tolower(adrs_geocoded$matched_city))

# Obtain which addresses have unmatched city
false_city <- which(!city_matching)



###### Check ZIPs
# Clean the input zip - remove excess whitespace (this is for the checking functions of zip code later)
cleaned_zip <- dht::clean_address(filled_zip)

# Check whether the matched zip code is the same as the zip code provided. Return the proportion of correctly matched zip code
zip_check <- function(zip, matched_zip) {
  check_zip <- zip ==  matched_zip
  proportion_matched <- mean(check_zip)
  return(proportion_matched)
}
zip_matched_proportion <- zip_check(tolower(cleaned_zip), tolower(adrs_geocoded$matched_zip))

# Obtain TRUE/FALSE indicating whether the geocoded zip code matched the input zip code
zip_match_status <- function(zip, matched_zip) {
  check_zip <- zip ==  matched_zip
  return(check_zip)
}
zip_matching <- zip_match_status(tolower(cleaned_zip), tolower(adrs_geocoded$matched_zip))

# Obtain which addresses have unmatched zip code
false_zip <- which(!zip_matching)



###### Check street names
# Street name matching checking

# Less stringent checking function for geocoded matched street names
street_check <- function(street, matched_street) {
  # The words that are "meaningless" or generic for street names, which doesn't help for examining matching
  excluded_words <- c("Ln", "Ave", "St", "Way", "Rd", "Dr", "Hwy", "Blvd", "Pkwy")
  
  # Split the geocoded matched street name into a vector of words
  splited_mat_str <- unlist(strsplit(matched_street, " "))
  
  # Create the regular expression pattern for the words need to be excluded from the matched street names
  pattern_exclude <- paste0("\\b(", paste(excluded_words, collapse = "|"), ")\\b")
  
  # Remove the words in the exclusion list from the matched street names
  mat_str_modified <- gsub(pattern_exclude, "", matched_street, ignore.case = TRUE)
  
  # Split the modified matched street name into words
  splited_mat_str_modified <- unlist(strsplit(mat_str_modified, " "))
  
  # More stringent matching: all words except those excluded (in the excluded_words vector) need to be in the input street name
  pattern <- paste0("\\b(", paste(splited_mat_str_modified, collapse = "|"), ")\\b")
  result <- grepl(pattern, street, ignore.case = TRUE)
  
  return(result)
}


# More stringent checking function for geocoded matched street names
street_check_stringent <- function(street, matched_street) {
  # The words that are "meaningless" or generic for streat names, which doesn't help for examining matching
  excluded_words <- c("Ln", "Ave", "St", "Way", "Rd", "Dr", "Hwy", "Blvd", "Pkwy")
  
  # Split the geocoded matched street name into a vector of words
  splited_mat_str <- unlist(strsplit(matched_street, " "))
  
  # Create the regular expression pattern for the words need to be excluded from the matched street names
  pattern_exclude <- paste0("\\b(", paste(excluded_words, collapse = "|"), ")\\b")
  
  # Remove the words in the exclusion list from the matched street names
  mat_str_modified <- gsub(pattern_exclude, "", matched_street, ignore.case = TRUE)
  
  # # Split the modified matched street name into words
  # splited_mat_str_modified <- unlist(strsplit(mat_str_modified, " "))
  
  # Remove the leading/trailing whitespace
  mat_str_modified2 <- trimws(mat_str_modified, "both")
  
  # More stringent matching: all words except those excluded (in the excluded_words vector) need to be in the input street name
  pattern <- paste0("\\b(", paste(mat_str_modified2, collapse = "|"), ")\\b")
  result <- grepl(pattern, street, ignore.case = TRUE)
  
  return(result)
}


# Check all the geocoded street names by using the more stringent street name checking function
street_check_list_stringent <- c()
for (i in 1:num_rows) {
  if (is.na(adrs_geocoded$matched_street[i])) {
    street_check_list_stringent <- c(street_check_list_stringent, NA)
  } else {
    current <- street_check_stringent(input_address$Street.Name[i], adrs_geocoded$matched_street[i])
    street_check_list_stringent <- c(street_check_list_stringent, current)
  }
}
# Obtain which addresses have unmatched street names
false_street_stringent <- which(!street_check_list_stringent)
# Obtain the proportion of matched streets (checked by stringent method)
street_stringent_matched_proportion <- mean(street_check_list_stringent)

# Check all the geocoded street names by using the less stringent street name checking function
street_check_list <- c()
for (i in 1:num_rows) {
  if (is.na(adrs_geocoded$matched_street[i])) {
    street_check_list <- c(street_check_list, NA)
  } else {
    current <- street_check(input_address$Street.Name[i], adrs_geocoded$matched_street[i])
    street_check_list <- c(street_check_list, current)
  }
}
# Obtain which addresses have unmatched street names
false_street <- which(!street_check_list)
# Obtain the proportion of matched streets (checked by less stringent method)
street_matched_proportion <- mean(street_check_list)



# Modify the adrs_geocoded to include the binary indicator
# (to indicate whether each of the component: state, city, zip code, and street name, have been correctly matched)
adrs_geocoded$street_matched <- street_check_list_stringent
adrs_geocoded$city_matched <- city_matching
adrs_geocoded$state_matched <- state_matching
adrs_geocoded$zip_matched <- zip_matching


# Add the date information (month, day, and year) to the adrs_geocoded
adrs_geocoded$month <- input_address$Month
adrs_geocoded$day <- input_address$Day
adrs_geocoded$year <- input_address$Year
adrs_geocoded$start <- input_address$start
adrs_geocoded$end <- input_address$end
adrs_geocoded$record_date <- input_address$record_date

# Modify the adrs_geocoded to include the ZIP codes from the original input addresses
adrs_geocoded$ZIP <- cleaned_zip

# Arrange the data by patient_id, address, year, and start date
adrs_geocoded <- adrs_geocoded %>%
  arrange(cohortPatientId, address, year, start)

# Function to combine periods for those patients report same addresses at different visits
combine <- function(data) {
  data <- data %>%
    group_by(cohortPatientId, address, year) %>%
    mutate(end = if_else(row_number() == 1, max(end), NA_Date_)) %>%
    filter(row_number() == 1) %>%
    ungroup()
  return(data)
}

# Apply the function to combine periods
adrs_geocoded <- combine(adrs_geocoded)

# Write the geocoded address into a csv file and save it to local folder
write.csv(adrs_geocoded, file = "adrs_geocoded.csv", row.names = FALSE)


# Output a txt file to report where the potential mistakes are
text_content <- list("The imprecisely geocoded addresses (without geocoded latitude or longitude): ")
if (imprecise_proportion != 1) {
  text_content <- c(text_content, paste(imprecise_geocoded, collapse = " "), "")
} else {
  text_content <- c(text_content, "No imprecisely geocoded addresses", "")
}

text_content <- c(text_content, "Unmatched states: ")
if (state_matched_proportion != 1) {
  text_content <- c(text_content, paste(false_state, collapse = " "), "")
} else {
  text_content <- c(text_content, "No unmatched state", "")
}

text_content <- c(text_content, "Unmatched cities: ")
if (city_matched_proportion != 1) {
  text_content <- c(text_content, paste(false_city, collapse = " "), "")
} else {
  text_content <- c(text_content, "No unmatched city", "")
}

text_content <- c(text_content, "Unmatched zip codes: ")
if (zip_matched_proportion != 1) {
  text_content <- c(text_content, paste(false_zip, collapse = " "), "")
} else {
  text_content <- c(text_content, "No unmatched zip code", "")
}

text_content <- c(text_content, "Unmatched street names: ")
if (street_stringent_matched_proportion != 1) {
  text_content <- c(text_content, paste(false_street_stringent, collapse = " "), "")
} else {
  text_content <- c(text_content, "No unmatched street name", "")
}

text_content <- c(text_content, "The numbers above correspond to the line number (Not \"ID\" number). Please go back and check whether there is any typo or wrong address. Once the potential mistakes have been corrected, input the fixed addresses and start over again." )

text_content <- unlist(text_content)
writeLines(text_content, "report.txt")

